@@ -21,7 +21,6 @@
 #include <sys/types.h>
 #include "queue.h"
 
-#include <assert.h>
 #include <err.h>
 #include <errno.h>
 #include <limits.h>
@@ -34,14 +33,121 @@
 
 static struct client_ctx	*client_next(struct client_ctx *);
 static struct client_ctx	*client_prev(struct client_ctx *);
-static void			 client_mtf(struct client_ctx *);
-static void			 client_none(struct screen_ctx *);
 static void			 client_placecalc(struct client_ctx *);
 static void			 client_wm_protocols(struct client_ctx *);
 static void			 client_mwm_hints(struct client_ctx *);
-static int			 client_inbound(struct client_ctx *, int, int);
 
-struct client_ctx	*curcc = NULL;
+struct client_ctx *
+client_init(Window win, struct screen_ctx *sc, int active)
+{
+	struct client_ctx	*cc;
+	XWindowAttributes	 wattr;
+	int			 mapped;
+	Window			 rwin, cwin;
+	int			 x, y, wx, wy;
+	unsigned int		 mask;
+
+	if (win == None)
+		return(NULL);
+	if (!XGetWindowAttributes(X_Dpy, win, &wattr))
+		return(NULL);
+
+	if (sc == NULL) {
+		if ((sc = screen_find(wattr.root)) == NULL)
+			return(NULL);
+		mapped = 1;
+	} else {
+		if (wattr.override_redirect || wattr.map_state != IsViewable)
+			return(NULL);
+		mapped = wattr.map_state != IsUnmapped;
+	}
+
+	cc = xmalloc(sizeof(*cc));
+
+	XGrabServer(X_Dpy);
+
+	cc->sc = sc;
+	cc->win = win;
+	cc->label = NULL;
+	cc->gc = NULL;
+	cc->flags = 0;
+	cc->stackingorder = 0;
+	memset(&cc->hint, 0, sizeof(cc->hint));
+	memset(&cc->ch, 0, sizeof(cc->ch));
+
+	TAILQ_INIT(&cc->nameq);
+	client_setname(cc);
+
+	conf_client(cc);
+
+	XGetClassHint(X_Dpy, cc->win, &cc->ch);
+	client_wm_hints(cc);
+	client_wm_protocols(cc);
+	client_getsizehints(cc);
+	client_mwm_hints(cc);
+
+	cc->geom.x = wattr.x;
+	cc->geom.y = wattr.y;
+	cc->geom.w = wattr.width;
+	cc->geom.h = wattr.height;
+	cc->dim.w = (cc->geom.w - cc->hint.basew) / cc->hint.incw;
+	cc->dim.h = (cc->geom.h - cc->hint.baseh) / cc->hint.inch;
+	cc->ptr.x = cc->geom.w / 2;
+	cc->ptr.y = cc->geom.h / 2;
+
+	cc->colormap = wattr.colormap;
+
+	if (wattr.map_state != IsViewable) {
+		client_placecalc(cc);
+		client_resize(cc, 0);
+		if ((cc->wmh) && (cc->wmh->flags & StateHint))
+			client_set_wm_state(cc, cc->wmh->initial_state);
+	} else {
+		if ((active == 0) && (XQueryPointer(X_Dpy, cc->win, &rwin,
+		    &cwin, &x, &y, &wx, &wy, &mask)) && (cwin != None))
+			active = 1;
+	}
+
+	XSelectInput(X_Dpy, cc->win, ColormapChangeMask | EnterWindowMask |
+	    PropertyChangeMask | KeyReleaseMask);
+
+	XAddToSaveSet(X_Dpy, cc->win);
+
+	client_transient(cc);
+
+	/* Notify client of its configuration. */
+	client_config(cc);
+
+	TAILQ_INSERT_TAIL(&sc->clientq, cc, entry);
+
+	xu_ewmh_net_client_list(sc);
+	xu_ewmh_net_client_list_stacking(sc);
+	xu_ewmh_restore_net_wm_state(cc);
+
+	if (client_get_wm_state(cc) == IconicState)
+		client_hide(cc);
+	else
+		client_unhide(cc);
+
+	if (mapped) {
+		if (group_restore(cc))
+			goto out;
+		if (group_autogroup(cc))
+			goto out;
+		if (Conf.stickygroups)
+			group_assign(sc->group_active, cc);
+		else
+			group_assign(NULL, cc);
+	}
+out:
+	XSync(X_Dpy, False);
+	XUngrabServer(X_Dpy);
+
+	if (active)
+		client_setactive(cc);
+
+	return(cc);
+}
 
 struct client_ctx *
 client_find(Window win)
@@ -58,91 +164,6 @@ client_find(Window win)
 	return(NULL);
 }
 
-struct client_ctx *
-client_init(Window win, struct screen_ctx *sc)
-{
-	struct client_ctx	*cc;
-	XWindowAttributes	 wattr;
-	int			 mapped;
-
-	if (win == None)
-		return(NULL);
-	if (!XGetWindowAttributes(X_Dpy, win, &wattr))
-		return(NULL);
-
-	if (sc == NULL) {
-		sc = screen_find(wattr.root);
-		mapped = 1;
-	} else {
-		if (wattr.override_redirect || wattr.map_state != IsViewable)
-			return(NULL);
-		mapped = wattr.map_state != IsUnmapped;
-	}
-
-	cc = xcalloc(1, sizeof(*cc));
-
-	XGrabServer(X_Dpy);
-
-	cc->sc = sc;
-	cc->win = win;
-
-	TAILQ_INIT(&cc->nameq);
-	client_setname(cc);
-
-	conf_client(cc);
-
-	XGetClassHint(X_Dpy, cc->win, &cc->ch);
-	client_wm_hints(cc);
-	client_wm_protocols(cc);
-	client_getsizehints(cc);
-	client_mwm_hints(cc);
-
-	/* Saved pointer position */
-	cc->ptr.x = -1;
-	cc->ptr.y = -1;
-
-	cc->geom.x = wattr.x;
-	cc->geom.y = wattr.y;
-	cc->geom.w = wattr.width;
-	cc->geom.h = wattr.height;
-	cc->colormap = wattr.colormap;
-
-	if (wattr.map_state != IsViewable) {
-		client_placecalc(cc);
-		client_move(cc);
-		if ((cc->wmh) && (cc->wmh->flags & StateHint))
-			client_set_wm_state(cc, cc->wmh->initial_state);
-	}
-
-	XSelectInput(X_Dpy, cc->win, ColormapChangeMask | EnterWindowMask |
-	    PropertyChangeMask | KeyReleaseMask);
-
-	XAddToSaveSet(X_Dpy, cc->win);
-
-	client_transient(cc);
-
-	/* Notify client of its configuration. */
-	client_config(cc);
-
-	TAILQ_INSERT_TAIL(&sc->clientq, cc, entry);
-
-	xu_ewmh_net_client_list(sc);
-	xu_ewmh_restore_net_wm_state(cc);
-
-	if (client_get_wm_state(cc) == IconicState)
-		client_hide(cc);
-	else
-		client_unhide(cc);
-
-	if (mapped)
-		group_autogroup(cc);
-
-	XSync(X_Dpy, False);
-	XUngrabServer(X_Dpy);
-
-	return(cc);
-}
-
 void
 client_delete(struct client_ctx *cc)
 {
@@ -152,12 +173,13 @@ client_delete(struct client_ctx *cc)
 	TAILQ_REMOVE(&sc->clientq, cc, entry);
 
 	xu_ewmh_net_client_list(sc);
+	xu_ewmh_net_client_list_stacking(sc);
 
-	if (cc->group != NULL)
-		TAILQ_REMOVE(&cc->group->clientq, cc, group_entry);
+	if (cc->flags & CLIENT_ACTIVE)
+		xu_ewmh_net_active_window(sc, None);
 
-	if (cc == client_current())
-		client_none(sc);
+	if (cc->gc != NULL)
+		TAILQ_REMOVE(&cc->gc->clientq, cc, group_entry);
 
 	while ((wn = TAILQ_FIRST(&cc->nameq)) != NULL) {
 		TAILQ_REMOVE(&cc->nameq, wn, entry);
@@ -194,16 +216,15 @@ client_setactive(struct client_ctx *cc)
 	if (cc->flags & CLIENT_WM_TAKE_FOCUS)
 		client_msg(cc, cwmh[WM_TAKE_FOCUS], Last_Event_Time);
 
-	if ((oldcc = client_current())) {
+	if ((oldcc = client_current()) != NULL) {
 		oldcc->flags &= ~CLIENT_ACTIVE;
 		client_draw_border(oldcc);
 	}
 
-	/* If we're in the middle of cycing, don't change the order. */
+	/* If we're in the middle of cycling, don't change the order. */
 	if (!sc->cycling)
 		client_mtf(cc);
 
-	curcc = cc;
 	cc->flags |= CLIENT_ACTIVE;
 	cc->flags &= ~CLIENT_URGENCY;
 	client_draw_border(cc);
@@ -211,53 +232,56 @@ client_setactive(struct client_ctx *cc)
 	xu_ewmh_net_active_window(sc, cc->win);
 }
 
-/*
- * set when there is no active client
- */
-static void
-client_none(struct screen_ctx *sc)
-{
-	Window none = None;
-
-	xu_ewmh_net_active_window(sc, none);
-
-	curcc = NULL;
-}
-
 struct client_ctx *
 client_current(void)
 {
-	return(curcc);
+	struct screen_ctx	*sc;
+	struct client_ctx	*cc;
+
+	TAILQ_FOREACH(sc, &Screenq, entry) {
+		TAILQ_FOREACH(cc, &sc->clientq, entry) {
+			if (cc->flags & CLIENT_ACTIVE)
+				return(cc);
+		}
+	}
+	return(NULL);
 }
 
 void
 client_toggle_freeze(struct client_ctx *cc)
 {
-	if (cc->flags & CLIENT_FREEZE)
-		cc->flags &= ~CLIENT_FREEZE;
-	else
-		cc->flags |= CLIENT_FREEZE;
+	if (cc->flags & CLIENT_FULLSCREEN)
+		return;
+
+	cc->flags ^= CLIENT_FREEZE;
+	xu_ewmh_set_net_wm_state(cc);
 }
 
 void
 client_toggle_hidden(struct client_ctx *cc)
 {
-	if (cc->flags & CLIENT_HIDDEN)
-		cc->flags &= ~CLIENT_HIDDEN;
-	else
-		cc->flags |= CLIENT_HIDDEN;
+	cc->flags ^= CLIENT_HIDDEN;
+	xu_ewmh_set_net_wm_state(cc);
+}
 
+void
+client_toggle_skip_pager(struct client_ctx *cc)
+{
+	cc->flags ^= CLIENT_SKIP_PAGER;
+	xu_ewmh_set_net_wm_state(cc);
+}
+
+void
+client_toggle_skip_taskbar(struct client_ctx *cc)
+{
+	cc->flags ^= CLIENT_SKIP_TASKBAR;
 	xu_ewmh_set_net_wm_state(cc);
 }
 
 void
 client_toggle_sticky(struct client_ctx *cc)
 {
-	if (cc->flags & CLIENT_STICKY)
-		cc->flags &= ~CLIENT_STICKY;
-	else
-		cc->flags |= CLIENT_STICKY;
-
+	cc->flags ^= CLIENT_STICKY;
 	xu_ewmh_set_net_wm_state(cc);
 }
 
@@ -265,14 +289,15 @@ void
 client_toggle_fullscreen(struct client_ctx *cc)
 {
 	struct screen_ctx	*sc = cc->sc;
-	struct geom		 xine;
+	struct geom		 area;
 
 	if ((cc->flags & CLIENT_FREEZE) &&
 	    !(cc->flags & CLIENT_FULLSCREEN))
 		return;
 
 	if (cc->flags & CLIENT_FULLSCREEN) {
-		cc->bwidth = Conf.bwidth;
+		if (!(cc->flags & CLIENT_IGNORE))
+			cc->bwidth = Conf.bwidth;
 		cc->geom = cc->fullgeom;
 		cc->flags &= ~(CLIENT_FULLSCREEN | CLIENT_FREEZE);
 		goto resize;
@@ -280,12 +305,12 @@ client_toggle_fullscreen(struct client_ctx *cc)
 
 	cc->fullgeom = cc->geom;
 
-	xine = screen_find_xinerama(sc,
+	area = screen_area(sc,
 	    cc->geom.x + cc->geom.w / 2,
 	    cc->geom.y + cc->geom.h / 2, CWM_NOGAP);
 
 	cc->bwidth = 0;
-	cc->geom = xine;
+	cc->geom = area;
 	cc->flags |= (CLIENT_FULLSCREEN | CLIENT_FREEZE);
 
 resize:
@@ -297,9 +322,9 @@ void
 client_toggle_maximize(struct client_ctx *cc)
 {
 	struct screen_ctx	*sc = cc->sc;
-	struct geom		 xine;
+	struct geom		 area;
 
-	if (cc->flags & (CLIENT_FREEZE|CLIENT_STICKY))
+	if (cc->flags & CLIENT_FREEZE)
 		return;
 
 	if ((cc->flags & CLIENT_MAXFLAGS) == CLIENT_MAXIMIZED) {
@@ -318,19 +343,14 @@ client_toggle_maximize(struct client_ctx *cc)
 		cc->savegeom.x = cc->geom.x;
 	}
 
-	/*
-	 * pick screen that the middle of the window is on.
-	 * that's probably more fair than if just the origin of
-	 * a window is poking over a boundary
-	 */
-	xine = screen_find_xinerama(sc,
+	area = screen_area(sc,
 	    cc->geom.x + cc->geom.w / 2,
 	    cc->geom.y + cc->geom.h / 2, CWM_GAP);
 
-	cc->geom.x = xine.x;
-	cc->geom.y = xine.y;
-	cc->geom.w = xine.w - (cc->bwidth * 2);
-	cc->geom.h = xine.h - (cc->bwidth * 2);
+	cc->geom.x = area.x;
+	cc->geom.y = area.y;
+	cc->geom.w = area.w - (cc->bwidth * 2);
+	cc->geom.h = area.h - (cc->bwidth * 2);
 	cc->flags |= CLIENT_MAXIMIZED;
 
 resize:
@@ -342,9 +362,9 @@ void
 client_toggle_vmaximize(struct client_ctx *cc)
 {
 	struct screen_ctx	*sc = cc->sc;
-	struct geom		 xine;
+	struct geom		 area;
 
-	if (cc->flags & (CLIENT_FREEZE|CLIENT_STICKY))
+	if (cc->flags & CLIENT_FREEZE)
 		return;
 
 	if (cc->flags & CLIENT_VMAXIMIZED) {
@@ -357,12 +377,12 @@ client_toggle_vmaximize(struct client_ctx *cc)
 	cc->savegeom.y = cc->geom.y;
 	cc->savegeom.h = cc->geom.h;
 
-	xine = screen_find_xinerama(sc,
+	area = screen_area(sc,
 	    cc->geom.x + cc->geom.w / 2,
 	    cc->geom.y + cc->geom.h / 2, CWM_GAP);
 
-	cc->geom.y = xine.y;
-	cc->geom.h = xine.h - (cc->bwidth * 2);
+	cc->geom.y = area.y;
+	cc->geom.h = area.h - (cc->bwidth * 2);
 	cc->flags |= CLIENT_VMAXIMIZED;
 
 resize:
@@ -374,9 +394,9 @@ void
 client_toggle_hmaximize(struct client_ctx *cc)
 {
 	struct screen_ctx	*sc = cc->sc;
-	struct geom		 xine;
+	struct geom		 area;
 
-	if (cc->flags & (CLIENT_FREEZE|CLIENT_STICKY))
+	if (cc->flags & CLIENT_FREEZE)
 		return;
 
 	if (cc->flags & CLIENT_HMAXIMIZED) {
@@ -389,12 +409,12 @@ client_toggle_hmaximize(struct client_ctx *cc)
 	cc->savegeom.x = cc->geom.x;
 	cc->savegeom.w = cc->geom.w;
 
-	xine = screen_find_xinerama(sc,
+	area = screen_area(sc,
 	    cc->geom.x + cc->geom.w / 2,
 	    cc->geom.y + cc->geom.h / 2, CWM_GAP);
 
-	cc->geom.x = xine.x;
-	cc->geom.w = xine.w - (cc->bwidth * 2);
+	cc->geom.x = area.x;
+	cc->geom.w = area.w - (cc->bwidth * 2);
 	cc->flags |= CLIENT_HMAXIMIZED;
 
 resize:
@@ -414,6 +434,8 @@ client_resize(struct client_ctx *cc, int reset)
 
 	XMoveResizeWindow(X_Dpy, cc->win, cc->geom.x,
 	    cc->geom.y, cc->geom.w, cc->geom.h);
+	cc->dim.w = (cc->geom.w - cc->hint.basew) / cc->hint.incw;
+	cc->dim.h = (cc->geom.h - cc->hint.baseh) / cc->hint.inch;
 	client_config(cc);
 }
 
@@ -456,21 +478,28 @@ client_config(struct client_ctx *cc)
 	XSendEvent(X_Dpy, cc->win, False, StructureNotifyMask, (XEvent *)&cn);
 }
 
+void
+client_ptr_inbound(struct client_ctx *cc, int getpos)
+{
+	if (getpos)
+		xu_ptr_getpos(cc->win, &cc->ptr.x, &cc->ptr.y);
+
+	if (cc->ptr.x < 0)
+		cc->ptr.x = 0;
+	else if (cc->ptr.x > cc->geom.w - 1)
+		cc->ptr.x = cc->geom.w - 1;
+	if (cc->ptr.y < 0)
+		cc->ptr.y = 0;
+	else if (cc->ptr.y > cc->geom.h - 1)
+		cc->ptr.y = cc->geom.h - 1;
+
+	client_ptrwarp(cc);
+}
+
 void
 client_ptrwarp(struct client_ctx *cc)
 {
-	int	 x = cc->ptr.x, y = cc->ptr.y;
-
-	if (x == -1 || y == -1) {
-		x = cc->geom.w / 2;
-		y = cc->geom.h / 2;
-	}
-
-	if (cc->flags & CLIENT_HIDDEN)
-		client_unhide(cc);
-	else
-		client_raise(cc);
-	xu_ptr_setpos(cc->win, x, y);
+	xu_ptr_setpos(cc->win, cc->ptr.x, cc->ptr.y);
 }
 
 void
@@ -483,33 +512,36 @@ client_ptrsave(struct client_ctx *cc)
 		cc->ptr.x = x;
 		cc->ptr.y = y;
 	} else {
-		cc->ptr.x = -1;
-		cc->ptr.y = -1;
+		cc->ptr.x = cc->geom.w / 2;
+		cc->ptr.y = cc->geom.h / 2;
 	}
 }
 
 void
 client_hide(struct client_ctx *cc)
 {
-	if (cc->flags & CLIENT_STICKY)
-		return;
-
 	XUnmapWindow(X_Dpy, cc->win);
 
+	if (cc->flags & CLIENT_ACTIVE)
+		xu_ewmh_net_active_window(cc->sc, None);
+
 	cc->flags &= ~CLIENT_ACTIVE;
 	cc->flags |= CLIENT_HIDDEN;
 	client_set_wm_state(cc, IconicState);
+}
 
-	if (cc == client_current())
-		client_none(cc->sc);
+void
+client_show(struct client_ctx *cc)
+{
+	if (cc->flags & CLIENT_HIDDEN)
+		client_unhide(cc);
+	else
+		client_raise(cc);
 }
 
 void
 client_unhide(struct client_ctx *cc)
 {
-	if (cc->flags & CLIENT_STICKY)
-		return;
-
 	XMapRaised(X_Dpy, cc->win);
 
 	cc->flags &= ~CLIENT_HIDDEN;
@@ -548,7 +580,7 @@ client_draw_border(struct client_ctx *cc)
 	if (cc->flags & CLIENT_URGENCY)
 		pixel = sc->xftcolor[CWM_COLOR_BORDER_URGENCY].pixel;
 
-	XSetWindowBorderWidth(X_Dpy, cc->win, cc->bwidth);
+	XSetWindowBorderWidth(X_Dpy, cc->win, (unsigned int)cc->bwidth);
 	XSetWindowBorder(X_Dpy, cc->win, pixel);
 }
 
@@ -612,6 +644,7 @@ client_setname(struct client_ctx *cc)
 {
 	struct winname	*wn;
 	char		*newname;
+	int		 i = 0;
 
 	if (!xu_getstrprop(cc->win, ewmh[_NET_WM_NAME], &newname))
 		if (!xu_getstrprop(cc->win, XA_WM_NAME, &newname))
@@ -628,97 +661,88 @@ client_setname(struct client_ctx *cc)
 	wn = xmalloc(sizeof(*wn));
 	wn->name = newname;
 	TAILQ_INSERT_TAIL(&cc->nameq, wn, entry);
-	cc->nameqlen++;
-
 match:
 	cc->name = wn->name;
 
-	/* Now, do some garbage collection. */
-	if (cc->nameqlen > CLIENT_MAXNAMEQLEN) {
+	/* Do some garbage collection. */
+	TAILQ_FOREACH(wn, &cc->nameq, entry)
+		i++;
+	if (i > Conf.nameqlen) {
 		wn = TAILQ_FIRST(&cc->nameq);
-		assert(wn != NULL);
 		TAILQ_REMOVE(&cc->nameq, wn, entry);
 		free(wn->name);
 		free(wn);
-		cc->nameqlen--;
+		i--;
 	}
 }
 
 void
 client_cycle(struct screen_ctx *sc, int flags)
 {
-	struct client_ctx	*oldcc, *newcc;
+	struct client_ctx	*newcc, *oldcc, *prevcc;
 	int			 again = 1;
 
 	if (TAILQ_EMPTY(&sc->clientq))
 		return;
 
+	prevcc = TAILQ_FIRST(&sc->clientq);
 	oldcc = client_current();
 	if (oldcc == NULL)
-		oldcc = (flags & CWM_RCYCLE ?
-		    TAILQ_LAST(&sc->clientq, client_ctx_q) :
-		    TAILQ_FIRST(&sc->clientq));
+		oldcc = (flags & CWM_CYCLE_REVERSE) ?
+		    TAILQ_LAST(&sc->clientq, client_q) :
+		    TAILQ_FIRST(&sc->clientq);
 
 	newcc = oldcc;
 	while (again) {
 		again = 0;
 
-		newcc = (flags & CWM_RCYCLE ? client_prev(newcc) :
-		    client_next(newcc));
+		newcc = (flags & CWM_CYCLE_REVERSE) ? client_prev(newcc) :
+		    client_next(newcc);
 
 		/* Only cycle visible and non-ignored windows. */
-		if ((newcc->flags & (CLIENT_HIDDEN|CLIENT_IGNORE))
-		    || ((flags & CWM_INGROUP) &&
-			(newcc->group != oldcc->group)))
+		if ((newcc->flags & (CLIENT_SKIP_CYCLE)) ||
+		    ((flags & CWM_CYCLE_INGROUP) &&
+		    (newcc->gc != oldcc->gc)))
 			again = 1;
 
 		/* Is oldcc the only non-hidden window? */
 		if (newcc == oldcc) {
 			if (again)
 				return;	/* No windows visible. */
-
 			break;
 		}
 	}
 
-	/* reset when cycling mod is released. XXX I hate this hack */
+	/* Reset when cycling mod is released. XXX I hate this hack */
 	sc->cycling = 1;
 	client_ptrsave(oldcc);
-	client_ptrwarp(newcc);
-}
-
-void
-client_cycle_leave(struct screen_ctx *sc)
-{
-	struct client_ctx	*cc;
-
-	sc->cycling = 0;
-
-	if ((cc = client_current())) {
-		client_mtf(cc);
-		group_toggle_membership_leave(cc);
-		XUngrabKeyboard(X_Dpy, CurrentTime);
+	client_raise(prevcc);
+	client_raise(newcc);
+	if (!client_inbound(newcc, newcc->ptr.x, newcc->ptr.y)) {
+		newcc->ptr.x = newcc->geom.w / 2;
+		newcc->ptr.y = newcc->geom.h / 2;
 	}
+	client_ptrwarp(newcc);
 }
 
 static struct client_ctx *
 client_next(struct client_ctx *cc)
 {
 	struct screen_ctx	*sc = cc->sc;
-	struct client_ctx	*ccc;
+	struct client_ctx	*newcc;
 
-	return((ccc = TAILQ_NEXT(cc, entry)) != NULL ?
-	    ccc : TAILQ_FIRST(&sc->clientq));
+	return(((newcc = TAILQ_NEXT(cc, entry)) != NULL) ?
+	    newcc : TAILQ_FIRST(&sc->clientq));
 }
 
 static struct client_ctx *
 client_prev(struct client_ctx *cc)
 {
 	struct screen_ctx	*sc = cc->sc;
-	struct client_ctx	*ccc;
+	struct client_ctx	*newcc;
 
-	return((ccc = TAILQ_PREV(cc, client_ctx_q, entry)) != NULL ?
-	    ccc : TAILQ_LAST(&sc->clientq, client_ctx_q));
+	return(((newcc = TAILQ_PREV(cc, client_q, entry)) != NULL) ?
+	    newcc : TAILQ_LAST(&sc->clientq, client_q));
 }
 
 static void
@@ -727,51 +751,48 @@ client_placecalc(struct client_ctx *cc)
 	struct screen_ctx	*sc = cc->sc;
 	int			 xslack, yslack;
 
-	if (cc->hint.flags & (USPosition|PPosition)) {
-		/*
-		 * Ignore XINERAMA screens, just make sure it's somewhere
-		 * in the virtual desktop. else it stops people putting xterms
-		 * at startup in the screen the mouse doesn't start in *sigh*.
-		 * XRandR bits mean that {x,y}max shouldn't be outside what's
-		 * currently there.
-		 */
-		xslack = sc->view.w - cc->geom.w - cc->bwidth * 2;
-		yslack = sc->view.h - cc->geom.h - cc->bwidth * 2;
-		cc->geom.x = MIN(cc->geom.x, xslack);
-		cc->geom.y = MIN(cc->geom.y, yslack);
+	if (cc->hint.flags & (USPosition | PPosition)) {
+		if (cc->geom.x >= sc->view.w)
+			cc->geom.x = sc->view.w - cc->bwidth - 1;
+		if (cc->geom.x + cc->geom.w + cc->bwidth <= 0)
+			cc->geom.x = -(cc->geom.w + cc->bwidth - 1);
+		if (cc->geom.y >= sc->view.h)
+			cc->geom.x = sc->view.h - cc->bwidth - 1;
+		if (cc->geom.y + cc->geom.h + cc->bwidth <= 0)
+			cc->geom.y = -(cc->geom.h + cc->bwidth - 1);
 	} else {
-		struct geom		 xine;
+		struct geom		 area;
 		int			 xmouse, ymouse;
 
 		xu_ptr_getpos(sc->rootwin, &xmouse, &ymouse);
-		xine = screen_find_xinerama(sc, xmouse, ymouse, CWM_GAP);
-		xine.w += xine.x;
-		xine.h += xine.y;
-		xmouse = MAX(xmouse, xine.x) - cc->geom.w / 2;
-		ymouse = MAX(ymouse, xine.y) - cc->geom.h / 2;
+		area = screen_area(sc, xmouse, ymouse, CWM_GAP);
+		area.w += area.x;
+		area.h += area.y;
+		xmouse = MAX(xmouse, area.x) - cc->geom.w / 2;
+		ymouse = MAX(ymouse, area.y) - cc->geom.h / 2;
 
-		xmouse = MAX(xmouse, xine.x);
-		ymouse = MAX(ymouse, xine.y);
+		xmouse = MAX(xmouse, area.x);
+		ymouse = MAX(ymouse, area.y);
 
-		xslack = xine.w - cc->geom.w - cc->bwidth * 2;
-		yslack = xine.h - cc->geom.h - cc->bwidth * 2;
+		xslack = area.w - cc->geom.w - cc->bwidth * 2;
+		yslack = area.h - cc->geom.h - cc->bwidth * 2;
 
-		if (xslack >= xine.x) {
-			cc->geom.x = MAX(MIN(xmouse, xslack), xine.x);
+		if (xslack >= area.x) {
+			cc->geom.x = MAX(MIN(xmouse, xslack), area.x);
 		} else {
-			cc->geom.x = xine.x;
-			cc->geom.w = xine.w;
+			cc->geom.x = area.x;
+			cc->geom.w = area.w;
 		}
-		if (yslack >= xine.y) {
-			cc->geom.y = MAX(MIN(ymouse, yslack), xine.y);
+		if (yslack >= area.y) {
+			cc->geom.y = MAX(MIN(ymouse, yslack), area.y);
 		} else {
-			cc->geom.y = xine.y;
-			cc->geom.h = xine.h;
+			cc->geom.y = area.y;
+			cc->geom.h = area.h;
 		}
 	}
 }
 
-static void
+void
 client_mtf(struct client_ctx *cc)
 {
 	struct screen_ctx	*sc = cc->sc;
@@ -815,6 +836,8 @@ client_getsizehints(struct client_ctx *cc)
 	}
 	cc->hint.incw = MAX(1, cc->hint.incw);
 	cc->hint.inch = MAX(1, cc->hint.inch);
+	cc->hint.minw = MAX(1, cc->hint.minw);
+	cc->hint.minh = MAX(1, cc->hint.minh);
 
 	if (size.flags & PAspect) {
 		if (size.min_aspect.x > 0)
@@ -871,9 +894,6 @@ client_applysizehints(struct client_ctx *cc)
 		cc->geom.w = MIN(cc->geom.w, cc->hint.maxw);
 	if (cc->hint.maxh)
 		cc->geom.h = MIN(cc->geom.h, cc->hint.maxh);
-
-	cc->geom.w = MAX(cc->geom.w, 1);
-	cc->geom.h = MAX(cc->geom.h, 1);
 }
 
 static void
@@ -881,12 +901,14 @@ client_mwm_hints(struct client_ctx *cc)
 {
 	struct mwm_hints	*mwmh;
 
-	if (xu_getprop(cc->win, cwmh[_MOTIF_WM_HINTS], cwmh[_MOTIF_WM_HINTS],
-	    PROP_MWM_HINTS_ELEMENTS, (unsigned char **)&mwmh) == MWM_NUMHINTS) {
-		if (mwmh->flags & MWM_HINTS_DECORATIONS &&
+	if (xu_getprop(cc->win, cwmh[_MOTIF_WM_HINTS],
+	    cwmh[_MOTIF_WM_HINTS], MWM_HINTS_ELEMENTS,
+	    (unsigned char **)&mwmh) == MWM_HINTS_ELEMENTS) {
+		if (mwmh->flags & MWM_FLAGS_DECORATIONS &&
 		    !(mwmh->decorations & MWM_DECOR_ALL) &&
 		    !(mwmh->decorations & MWM_DECOR_BORDER))
 			cc->bwidth = 0;
+		XFree(mwmh);
 	}
 }
 
@@ -897,15 +919,15 @@ client_transient(struct client_ctx *cc)
 	Window			 trans;
 
 	if (XGetTransientForHint(X_Dpy, cc->win, &trans)) {
-		if ((tc = client_find(trans)) && tc->group) {
-			group_movetogroup(cc, tc->group->num);
+		if ((tc = client_find(trans)) != NULL && tc->gc) {
+			group_movetogroup(cc, tc->gc->num);
 			if (tc->flags & CLIENT_IGNORE)
 				cc->flags |= CLIENT_IGNORE;
 		}
 	}
 }
 
-static int
+int
 client_inbound(struct client_ctx *cc, int x, int y)
 {
 	return(x < cc->geom.w && x >= 0 &&
@@ -943,10 +965,10 @@ void
 client_htile(struct client_ctx *cc)
 {
 	struct client_ctx	*ci;
-	struct group_ctx 	*gc = cc->group;
+	struct group_ctx 	*gc = cc->gc;
 	struct screen_ctx 	*sc = cc->sc;
-	struct geom 		 xine;
-	int 			 i, n, mh, x, h, w;
+	struct geom 		 area;
+	int 			 i, n, mh, x, w, h;
 
 	if (!gc)
 		return;
@@ -961,40 +983,41 @@ client_htile(struct client_ctx *cc)
 	if (n == 0)
 		return;
 
-	xine = screen_find_xinerama(sc,
+	area = screen_area(sc,
 	    cc->geom.x + cc->geom.w / 2,
 	    cc->geom.y + cc->geom.h / 2, CWM_GAP);
 
 	if (cc->flags & CLIENT_VMAXIMIZED ||
-	    cc->geom.h + (cc->bwidth * 2) >= xine.h)
+	    cc->geom.h + (cc->bwidth * 2) >= area.h)
 		return;
 
 	cc->flags &= ~CLIENT_HMAXIMIZED;
-	cc->geom.x = xine.x;
-	cc->geom.y = xine.y;
-	cc->geom.w = xine.w - (cc->bwidth * 2);
+	cc->geom.x = area.x;
+	cc->geom.y = area.y;
+	cc->geom.w = area.w - (cc->bwidth * 2);
+	cc->geom.h = (area.h - (cc->bwidth * 2)) / 2;
 	client_resize(cc, 1);
 	client_ptrwarp(cc);
 
 	mh = cc->geom.h + (cc->bwidth * 2);
-	x = xine.x;
-	w = xine.w / n;
-	h = xine.h - mh;
+	x = area.x;
+	w = area.w / n;
+	h = area.h - mh;
 	TAILQ_FOREACH(ci, &gc->clientq, group_entry) {
 		if (ci->flags & CLIENT_HIDDEN ||
 		    ci->flags & CLIENT_IGNORE || (ci == cc))
 			continue;
 		ci->bwidth = Conf.bwidth;
-		ci->geom.y = xine.y + mh;
 		ci->geom.x = x;
-		ci->geom.h = h - (ci->bwidth * 2);
+		ci->geom.y = area.y + mh;
 		ci->geom.w = w - (ci->bwidth * 2);
+		ci->geom.h = h - (ci->bwidth * 2);
 		if (i + 1 == n)
-			ci->geom.w = xine.x + xine.w -
+			ci->geom.w = area.x + area.w -
 			    ci->geom.x - (ci->bwidth * 2);
 		x += w;
-		client_resize(ci, 1);
 		i++;
+		client_resize(ci, 1);
 	}
 }
 
@@ -1002,10 +1025,10 @@ void
 client_vtile(struct client_ctx *cc)
 {
 	struct client_ctx	*ci;
-	struct group_ctx 	*gc = cc->group;
+	struct group_ctx 	*gc = cc->gc;
 	struct screen_ctx 	*sc = cc->sc;
-	struct geom 		 xine;
-	int 			 i, n, mw, y, h, w;
+	struct geom 		 area;
+	int 			 i, n, mw, y, w, h;
 
 	if (!gc)
 		return;
@@ -1020,40 +1043,41 @@ client_vtile(struct client_ctx *cc)
 	if (n == 0)
 		return;
 
-	xine = screen_find_xinerama(sc,
+	area = screen_area(sc,
 	    cc->geom.x + cc->geom.w / 2,
 	    cc->geom.y + cc->geom.h / 2, CWM_GAP);
 
 	if (cc->flags & CLIENT_HMAXIMIZED ||
-	    cc->geom.w + (cc->bwidth * 2) >= xine.w)
+	    cc->geom.w + (cc->bwidth * 2) >= area.w)
 		return;
 
 	cc->flags &= ~CLIENT_VMAXIMIZED;
-	cc->geom.x = xine.x;
-	cc->geom.y = xine.y;
-	cc->geom.h = xine.h - (cc->bwidth * 2);
+	cc->geom.x = area.x;
+	cc->geom.y = area.y;
+	cc->geom.w = (area.w - (cc->bwidth * 2)) / 2;
+	cc->geom.h = area.h - (cc->bwidth * 2);
 	client_resize(cc, 1);
 	client_ptrwarp(cc);
 
 	mw = cc->geom.w + (cc->bwidth * 2);
-	y = xine.y;
-	h = xine.h / n;
-	w = xine.w - mw;
+	y = area.y;
+	h = area.h / n;
+	w = area.w - mw;
 	TAILQ_FOREACH(ci, &gc->clientq, group_entry) {
 		if (ci->flags & CLIENT_HIDDEN ||
 		    ci->flags & CLIENT_IGNORE || (ci == cc))
 			continue;
 		ci->bwidth = Conf.bwidth;
+		ci->geom.x = area.x + mw;
 		ci->geom.y = y;
-		ci->geom.x = xine.x + mw;
-		ci->geom.h = h - (ci->bwidth * 2);
 		ci->geom.w = w - (ci->bwidth * 2);
+		ci->geom.h = h - (ci->bwidth * 2);
 		if (i + 1 == n)
-			ci->geom.h = xine.y + xine.h -
+			ci->geom.h = area.y + area.h -
 			    ci->geom.y - (ci->bwidth * 2);
 		y += h;
-		client_resize(ci, 1);
 		i++;
+		client_resize(ci, 1);
 	}
 }
 
@@ -1078,4 +1102,3 @@ client_set_wm_state(struct client_ctx *cc, long state)
 	XChangeProperty(X_Dpy, cc->win, cwmh[WM_STATE], cwmh[WM_STATE], 32,
 	    PropModeReplace, (unsigned char *)data, 2);
 }
-
