@@ -48,34 +48,29 @@ struct menu_ctx {
 	char			 searchstr[MENU_MAXENTRY + 1];
 	char			 dispstr[MENU_MAXENTRY*2 + 1];
 	char			 promptstr[MENU_MAXENTRY + 1];
-	int			 hasprompt;
 	int			 list;
 	int			 listing;
 	int			 changed;
-	int			 noresult;
 	int			 prev;
 	int			 entry;
-	int			 height;
-	int			 width;
 	int			 num;
 	int 			 flags;
-	int			 x;
-	int			 y;
+	struct geom		 geom;
 	void (*match)(struct menu_q *, struct menu_q *, char *);
 	void (*print)(struct menu *, int);
 };
 static struct menu	*menu_handle_key(XEvent *, struct menu_ctx *,
 			     struct menu_q *, struct menu_q *);
-static void		 menu_handle_move(XEvent *, struct menu_ctx *,
-			     struct menu_q *);
-static struct menu	*menu_handle_release(XEvent *, struct menu_ctx *,
-			     struct menu_q *);
+static void		 menu_handle_move(struct menu_ctx *,
+			     struct menu_q *, int, int);
+static struct menu	*menu_handle_release(struct menu_ctx *,
+			     struct menu_q *, int, int);
 static void		 menu_draw(struct menu_ctx *, struct menu_q *,
 			     struct menu_q *);
 static void 		 menu_draw_entry(struct menu_ctx *, struct menu_q *,
 			     int, int);
 static int		 menu_calc_entry(struct menu_ctx *, int, int);
-static struct menu 	*menu_complete_path(struct menu_ctx *);
+static struct menu	*menu_complete_path(struct menu_ctx *);
 static int		 menu_keycode(XKeyEvent *, enum ctltype *, char *);
 
 struct menu *
@@ -89,58 +84,51 @@ menu_filter(struct screen_ctx *sc, struct menu_q *menuq, const char *prompt,
 	struct menu		*mi = NULL;
 	XEvent			 e;
 	Window			 focuswin;
-	int			 evmask, focusrevert;
-	int			 xsave, ysave, xcur, ycur;
+	int			 focusrevert, xsave, ysave, xcur, ycur;
 
 	TAILQ_INIT(&resultq);
 
+	xu_ptr_getpos(sc->rootwin, &xsave, &ysave);
+
 	(void)memset(&mc, 0, sizeof(mc));
-
-	xu_ptr_getpos(sc->rootwin, &mc.x, &mc.y);
-
-	xsave = mc.x;
-	ysave = mc.y;
-
 	mc.sc = sc;
 	mc.flags = flags;
-	if (prompt != NULL) {
-		evmask = MENUMASK | KEYMASK; /* accept keys as well */
-		(void)strlcpy(mc.promptstr, prompt, sizeof(mc.promptstr));
-		mc.hasprompt = 1;
-	} else {
-		evmask = MENUMASK;
-		mc.list = 1;
-	}
+	mc.match = match;
+	mc.print = print;
+	mc.entry = mc.prev = -1;
+	mc.geom.x = xsave;
+	mc.geom.y = ysave;
 
+	if (mc.flags & CWM_MENU_LIST)
+		mc.list = 1;
+
+	(void)strlcpy(mc.promptstr, prompt, sizeof(mc.promptstr));
 	if (initial != NULL)
 		(void)strlcpy(mc.searchstr, initial, sizeof(mc.searchstr));
 	else
 		mc.searchstr[0] = '\0';
 
-	mc.match = match;
-	mc.print = print;
-	mc.entry = mc.prev = -1;
+	XSelectInput(X_Dpy, sc->menu.win, MENUMASK);
+	XMapRaised(X_Dpy, sc->menu.win);
 
-	XSelectInput(X_Dpy, sc->menuwin, evmask);
-	XMapRaised(X_Dpy, sc->menuwin);
-
-	if (xu_ptr_grab(sc->menuwin, MENUGRABMASK,
-	    Conf.cursor[CF_QUESTION]) < 0) {
-		XUnmapWindow(X_Dpy, sc->menuwin);
+	if (XGrabPointer(X_Dpy, sc->menu.win, False, MENUGRABMASK,
+	    GrabModeAsync, GrabModeAsync, None, Conf.cursor[CF_QUESTION],
+	    CurrentTime) != GrabSuccess) {
+		XUnmapWindow(X_Dpy, sc->menu.win);
 		return(NULL);
 	}
 
 	XGetInputFocus(X_Dpy, &focuswin, &focusrevert);
-	XSetInputFocus(X_Dpy, sc->menuwin, RevertToPointerRoot, CurrentTime);
+	XSetInputFocus(X_Dpy, sc->menu.win, RevertToPointerRoot, CurrentTime);
 
 	/* make sure keybindings don't remove keys from the menu stream */
-	XGrabKeyboard(X_Dpy, sc->menuwin, True,
+	XGrabKeyboard(X_Dpy, sc->menu.win, True,
 	    GrabModeAsync, GrabModeAsync, CurrentTime);
 
 	for (;;) {
 		mc.changed = 0;
 
-		XWindowEvent(X_Dpy, sc->menuwin, evmask, &e);
+		XWindowEvent(X_Dpy, sc->menu.win, MENUMASK, &e);
 
 		switch (e.type) {
 		case KeyPress:
@@ -152,11 +140,12 @@ menu_filter(struct screen_ctx *sc, struct menu_q *menuq, const char *prompt,
 			menu_draw(&mc, menuq, &resultq);
 			break;
 		case MotionNotify:
-			menu_handle_move(&e, &mc, &resultq);
+			menu_handle_move(&mc, &resultq,
+			    e.xbutton.x, e.xbutton.y);
 			break;
 		case ButtonRelease:
-			if ((mi = menu_handle_release(&e, &mc, &resultq))
-			    != NULL)
+			if ((mi = menu_handle_release(&mc, &resultq,
+			    e.xbutton.x, e.xbutton.y)) != NULL)
 				goto out;
 			break;
 		default:
@@ -173,12 +162,12 @@ out:
 	XSetInputFocus(X_Dpy, focuswin, focusrevert, CurrentTime);
 	/* restore if user didn't move */
 	xu_ptr_getpos(sc->rootwin, &xcur, &ycur);
-	if (xcur == mc.x && ycur == mc.y)
+	if (xcur == mc.geom.x && ycur == mc.geom.y)
 		xu_ptr_setpos(sc->rootwin, xsave, ysave);
-	xu_ptr_ungrab();
+	XUngrabPointer(X_Dpy, CurrentTime);
 
-	XMoveResizeWindow(X_Dpy, sc->menuwin, 0, 0, 1, 1);
-	XUnmapWindow(X_Dpy, sc->menuwin);
+	XMoveResizeWindow(X_Dpy, sc->menu.win, 0, 0, 1, 1);
+	XUnmapWindow(X_Dpy, sc->menu.win);
 	XUngrabKeyboard(X_Dpy, CurrentTime);
 
 	return(mi);
@@ -187,30 +176,28 @@ out:
 static struct menu *
 menu_complete_path(struct menu_ctx *mc)
 {
+	struct screen_ctx	*sc = mc->sc;
 	struct menu		*mi, *mr;
 	struct menu_q		 menuq;
-	char			*path = NULL;
+	int			 mflags = (CWM_MENU_DUMMY);
 
-	path = xcalloc(1, sizeof(mr->text));
 	mr = xcalloc(1, sizeof(*mr));
 
 	TAILQ_INIT(&menuq);
 
-	if ((mi = menu_filter(mc->sc, &menuq, mc->searchstr, NULL,
-	    CWM_MENU_DUMMY, search_match_path_any, NULL)) != NULL) {
+	if ((mi = menu_filter(sc, &menuq, mc->searchstr, NULL, mflags,
+	    search_match_path, search_print_text)) != NULL) {
 		mr->abort = mi->abort;
 		mr->dummy = mi->dummy;
-		strlcpy(path, mi->text, sizeof(mi->text));
+		if (mi->text[0] != '\0')
+			snprintf(mr->text, sizeof(mr->text), "%s \"%s\"",
+			    mc->searchstr, mi->text);
+		else if (!mr->abort)
+			strlcpy(mr->text, mc->searchstr, sizeof(mr->text));
 	}
 	
 	menuq_clear(&menuq);
 
-	if (path[0] != '\0')
-		snprintf(mr->text, sizeof(mr->text), "%s \"%s\"",
-			mc->searchstr, path);
-	else if (!mr->abort)
-		strlcpy(mr->text,  mc->searchstr, sizeof(mr->text));
-	free(path);
 	return(mr);
 }
 
@@ -317,15 +304,10 @@ menu_handle_key(XEvent *e, struct menu_ctx *mc, struct menu_q *menuq,
 		(void)strlcat(mc->searchstr, chr, sizeof(mc->searchstr));
 	}
 
-	mc->noresult = 0;
-	if (mc->changed && mc->searchstr[0] != '\0') {
-		(*mc->match)(menuq, resultq, mc->searchstr);
-		/* If menuq is empty, never show we've failed */
-		mc->noresult = TAILQ_EMPTY(resultq) && !TAILQ_EMPTY(menuq);
-	} else if (mc->changed)
-		TAILQ_INIT(resultq);
-
-	if (!mc->list && mc->listing && !mc->changed) {
+	if (mc->changed) {
+		if (mc->searchstr[0] != '\0')
+			(*mc->match)(menuq, resultq, mc->searchstr);
+	} else if (!mc->list && mc->listing) {
 		TAILQ_INIT(resultq);
 		mc->listing = 0;
 	}
@@ -338,98 +320,87 @@ menu_draw(struct menu_ctx *mc, struct menu_q *menuq, struct menu_q *resultq)
 {
 	struct screen_ctx	*sc = mc->sc;
 	struct menu		*mi;
-	struct geom		 xine;
+	struct geom		 area;
 	int			 n, xsave, ysave;
+	XGlyphInfo		 extents;
 
 	if (mc->list) {
-		if (TAILQ_EMPTY(resultq) && mc->list) {
+		if (TAILQ_EMPTY(resultq)) {
 			/* Copy them all over. */
 			TAILQ_FOREACH(mi, menuq, entry)
-				TAILQ_INSERT_TAIL(resultq, mi,
-				    resultentry);
+				TAILQ_INSERT_TAIL(resultq, mi, resultentry);
 
 			mc->listing = 1;
 		} else if (mc->changed)
 			mc->listing = 0;
 	}
 
-	mc->num = 0;
-	mc->width = 0;
-	mc->height = 0;
-	if (mc->hasprompt) {
-		(void)snprintf(mc->dispstr, sizeof(mc->dispstr), "%s%s%s%s",
-		    mc->promptstr, PROMPT_SCHAR, mc->searchstr, PROMPT_ECHAR);
-		mc->width = xu_xft_width(sc->xftfont, mc->dispstr,
-		    strlen(mc->dispstr));
-		mc->height = sc->xftfont->height + 1;
-		mc->num = 1;
-	}
+	(void)snprintf(mc->dispstr, sizeof(mc->dispstr), "%s%s%s%s",
+	    mc->promptstr, PROMPT_SCHAR, mc->searchstr, PROMPT_ECHAR);
+	XftTextExtentsUtf8(X_Dpy, sc->xftfont,
+	    (const FcChar8*)mc->dispstr, strlen(mc->dispstr), &extents);
+	mc->geom.w = extents.xOff;
+	mc->geom.h = sc->xftfont->height + 1;
+	mc->num = 1;
 
 	TAILQ_FOREACH(mi, resultq, resultentry) {
-		char *text;
-
-		if (mc->print != NULL) {
-			(*mc->print)(mi, mc->listing);
-			text = mi->print;
-		} else {
-			mi->print[0] = '\0';
-			text = mi->text;
-		}
-
-		mc->width = MAX(mc->width, xu_xft_width(sc->xftfont, text,
-		    MIN(strlen(text), MENU_MAXENTRY)));
-		mc->height += sc->xftfont->height + 1;
+		(*mc->print)(mi, mc->listing);
+		XftTextExtentsUtf8(X_Dpy, sc->xftfont,
+		    (const FcChar8*)mi->print,
+		    MIN(strlen(mi->print), MENU_MAXENTRY), &extents);
+		mc->geom.w = MAX(mc->geom.w, extents.xOff);
+		mc->geom.h += sc->xftfont->height + 1;
 		mc->num++;
 	}
 
-	xine = screen_find_xinerama(sc, mc->x, mc->y, CWM_GAP);
-	xine.w += xine.x - Conf.bwidth * 2;
-	xine.h += xine.y - Conf.bwidth * 2;
+	area = screen_area(sc, mc->geom.x, mc->geom.y, CWM_GAP);
+	area.w += area.x - Conf.bwidth * 2;
+	area.h += area.y - Conf.bwidth * 2;
 
-	xsave = mc->x;
-	ysave = mc->y;
+	xsave = mc->geom.x;
+	ysave = mc->geom.y;
 
 	/* Never hide the top, or left side, of the menu. */
-	if (mc->x + mc->width >= xine.w)
-		mc->x = xine.w - mc->width;
-	if (mc->x < xine.x) {
-		mc->x = xine.x;
-		mc->width = MIN(mc->width, (xine.w - xine.x));
+	if (mc->geom.x + mc->geom.w >= area.w)
+		mc->geom.x = area.w - mc->geom.w;
+	if (mc->geom.x < area.x) {
+		mc->geom.x = area.x;
+		mc->geom.w = MIN(mc->geom.w, (area.w - area.x));
 	}
-	if (mc->y + mc->height >= xine.h)
-		mc->y = xine.h - mc->height;
-	if (mc->y < xine.y) {
-		mc->y = xine.y;
-		mc->height = MIN(mc->height, (xine.h - xine.y));
+	if (mc->geom.y + mc->geom.h >= area.h)
+		mc->geom.y = area.h - mc->geom.h;
+	if (mc->geom.y < area.y) {
+		mc->geom.y = area.y;
+		mc->geom.h = MIN(mc->geom.h, (area.h - area.y));
 	}
 
-	if (mc->x != xsave || mc->y != ysave)
-		xu_ptr_setpos(sc->rootwin, mc->x, mc->y);
+	if (mc->geom.x != xsave || mc->geom.y != ysave)
+		xu_ptr_setpos(sc->rootwin, mc->geom.x, mc->geom.y);
 
-	XClearWindow(X_Dpy, sc->menuwin);
-	XMoveResizeWindow(X_Dpy, sc->menuwin, mc->x, mc->y,
-	    mc->width, mc->height);
+	XClearWindow(X_Dpy, sc->menu.win);
+	XMoveResizeWindow(X_Dpy, sc->menu.win, mc->geom.x, mc->geom.y,
+	    mc->geom.w, mc->geom.h);
 
-	if (mc->hasprompt) {
-		xu_xft_draw(sc, mc->dispstr, CWM_COLOR_MENU_FONT,
-		    0, sc->xftfont->ascent);
-		n = 1;
-	} else
-		n = 0;
+	n = 1;
+	XftDrawStringUtf8(sc->menu.xftdraw,
+	    &sc->xftcolor[CWM_COLOR_MENU_FONT], sc->xftfont,
+	    0, sc->xftfont->ascent,
+	    (const FcChar8*)mc->dispstr, strlen(mc->dispstr));
 
 	TAILQ_FOREACH(mi, resultq, resultentry) {
-		char *text = mi->print[0] != '\0' ?
-		    mi->print : mi->text;
 		int y = n * (sc->xftfont->height + 1) + sc->xftfont->ascent + 1;
 
 		/* Stop drawing when menu doesn't fit inside the screen. */
-		if (mc->y + y > xine.h)
+		if (mc->geom.y + y > area.h)
 			break;
 
-		xu_xft_draw(sc, text, CWM_COLOR_MENU_FONT, 0, y);
+		XftDrawStringUtf8(sc->menu.xftdraw,
+		    &sc->xftcolor[CWM_COLOR_MENU_FONT], sc->xftfont,
+		    0, y,
+		    (const FcChar8*)mi->print, strlen(mi->print));
 		n++;
 	}
-	if (mc->hasprompt && n > 1)
+	if (n > 1)
 		menu_draw_entry(mc, resultq, 1, 1);
 }
 
@@ -439,11 +410,7 @@ menu_draw_entry(struct menu_ctx *mc, struct menu_q *resultq,
 {
 	struct screen_ctx	*sc = mc->sc;
 	struct menu		*mi;
-	char 			*text;
-	int			 color, i = 0;
-
-	if (mc->hasprompt)
-		i = 1;
+	int			 color, i = 1;
 
 	TAILQ_FOREACH(mi, resultq, resultentry)
 		if (entry == i++)
@@ -451,21 +418,22 @@ menu_draw_entry(struct menu_ctx *mc, struct menu_q *resultq,
 	if (mi == NULL)
 		return;
 
-	color = active ? CWM_COLOR_MENU_FG : CWM_COLOR_MENU_BG;
-	text = mi->print[0] != '\0' ? mi->print : mi->text;
-	XftDrawRect(sc->xftdraw, &sc->xftcolor[color], 0,
-	    (sc->xftfont->height + 1) * entry, mc->width,
+	color = (active) ? CWM_COLOR_MENU_FG : CWM_COLOR_MENU_BG;
+	XftDrawRect(sc->menu.xftdraw, &sc->xftcolor[color], 0,
+	    (sc->xftfont->height + 1) * entry, mc->geom.w,
 	    (sc->xftfont->height + 1) + sc->xftfont->descent);
-	color = active ? CWM_COLOR_MENU_FONT_SEL : CWM_COLOR_MENU_FONT;
-	xu_xft_draw(sc, text, color,
-	    0, (sc->xftfont->height + 1) * entry + sc->xftfont->ascent + 1);
+	color = (active) ? CWM_COLOR_MENU_FONT_SEL : CWM_COLOR_MENU_FONT;
+	XftDrawStringUtf8(sc->menu.xftdraw,
+	    &sc->xftcolor[color], sc->xftfont,
+	    0, (sc->xftfont->height + 1) * entry + sc->xftfont->ascent + 1,
+	    (const FcChar8*)mi->print, strlen(mi->print));
 }
 
 static void
-menu_handle_move(XEvent *e, struct menu_ctx *mc, struct menu_q *resultq)
+menu_handle_move(struct menu_ctx *mc, struct menu_q *resultq, int x, int y)
 {
 	mc->prev = mc->entry;
-	mc->entry = menu_calc_entry(mc, e->xbutton.x, e->xbutton.y);
+	mc->entry = menu_calc_entry(mc, x, y);
 
 	if (mc->prev == mc->entry)
 		return;
@@ -473,22 +441,19 @@ menu_handle_move(XEvent *e, struct menu_ctx *mc, struct menu_q *resultq)
 	if (mc->prev != -1)
 		menu_draw_entry(mc, resultq, mc->prev, 0);
 	if (mc->entry != -1) {
-		(void)xu_ptr_regrab(MENUGRABMASK, Conf.cursor[CF_NORMAL]);
+		XChangeActivePointerGrab(X_Dpy, MENUGRABMASK,
+		    Conf.cursor[CF_NORMAL], CurrentTime);
 		menu_draw_entry(mc, resultq, mc->entry, 1);
-	} else
-		(void)xu_ptr_regrab(MENUGRABMASK, Conf.cursor[CF_DEFAULT]);
+	}
 }
 
 static struct menu *
-menu_handle_release(XEvent *e, struct menu_ctx *mc, struct menu_q *resultq)
+menu_handle_release(struct menu_ctx *mc, struct menu_q *resultq, int x, int y)
 {
 	struct menu		*mi;
-	int			 entry, i = 0;
+	int			 entry, i = 1;
 
-	entry = menu_calc_entry(mc, e->xbutton.x, e->xbutton.y);
-
-	if (mc->hasprompt)
-		i = 1;
+	entry = menu_calc_entry(mc, x, y);
 
 	TAILQ_FOREACH(mi, resultq, resultentry)
 		if (entry == i++)
@@ -510,12 +475,12 @@ menu_calc_entry(struct menu_ctx *mc, int x, int y)
 	entry = y / (sc->xftfont->height + 1);
 
 	/* in bounds? */
-	if (x < 0 || x > mc->width || y < 0 ||
+	if (x < 0 || x > mc->geom.w || y < 0 ||
 	    y > (sc->xftfont->height + 1) * mc->num ||
 	    entry < 0 || entry >= mc->num)
 		entry = -1;
 
-	if (mc->hasprompt && entry == 0)
+	if (entry == 0)
 		entry = -1;
 
 	return(entry);
@@ -525,19 +490,19 @@ static int
 menu_keycode(XKeyEvent *ev, enum ctltype *ctl, char *chr)
 {
 	KeySym		 ks;
-	unsigned int 	 state = ev->state;
 
 	*ctl = CTL_NONE;
 	chr[0] = '\0';
 
 	ks = XkbKeycodeToKeysym(X_Dpy, ev->keycode, 0,
-	     (state & ShiftMask) ? 1 : 0);
+	     (ev->state & ShiftMask) ? 1 : 0);
 
 	/* Look for control characters. */
 	switch (ks) {
 	case XK_BackSpace:
 		*ctl = CTL_ERASEONE;
 		break;
+	case XK_KP_Enter:
 	case XK_Return:
 		*ctl = CTL_RETURN;
 		break;
@@ -555,7 +520,7 @@ menu_keycode(XKeyEvent *ev, enum ctltype *ctl, char *chr)
 		break;
 	}
 
-	if (*ctl == CTL_NONE && (state & ControlMask)) {
+	if (*ctl == CTL_NONE && (ev->state & ControlMask)) {
 		switch (ks) {
 		case XK_s:
 		case XK_S:
@@ -579,10 +544,13 @@ menu_keycode(XKeyEvent *ev, enum ctltype *ctl, char *chr)
 		case XK_A:
 			*ctl = CTL_ALL;
 			break;
+		case XK_bracketleft:
+			*ctl = CTL_ABORT;
+			break;
 		}
 	}
 
-	if (*ctl == CTL_NONE && (state & Mod1Mask)) {
+	if (*ctl == CTL_NONE && (ev->state & Mod1Mask)) {
 		switch (ks) {
 		case XK_j:
 		case XK_J:
@@ -616,7 +584,10 @@ menuq_add(struct menu_q *mq, void *ctx, const char *fmt, ...)
 	mi->ctx = ctx;
 
 	va_start(ap, fmt);
-	(void)vsnprintf(mi->text, sizeof(mi->text), fmt, ap);
+	if (fmt != NULL)
+		(void)vsnprintf(mi->text, sizeof(mi->text), fmt, ap);
+	else
+		mi->text[0] = '\0';
 	va_end(ap);
 
 	TAILQ_INSERT_TAIL(mq, mi, entry);
@@ -632,3 +603,34 @@ menuq_clear(struct menu_q *mq)
 		free(mi);
 	}
 }
+
+void
+menu_windraw(struct screen_ctx *sc, Window win, const char *fmt, ...)
+{
+	va_list			 ap;
+	int			 i;
+	char			*text;
+	XGlyphInfo		 extents;
+
+	va_start(ap, fmt);
+	i = vasprintf(&text, fmt, ap);
+	va_end(ap);
+
+	if (i < 0 || text == NULL)
+		err(1, "vasprintf");
+
+	XftTextExtentsUtf8(X_Dpy, sc->xftfont, (const FcChar8*)text,
+	    strlen(text), &extents);
+
+	XReparentWindow(X_Dpy, sc->menu.win, win, 0, 0);
+	XMoveResizeWindow(X_Dpy, sc->menu.win, 0, 0,
+	    extents.xOff, sc->xftfont->height);
+	XMapWindow(X_Dpy, sc->menu.win);
+	XClearWindow(X_Dpy, sc->menu.win);
+
+	XftDrawStringUtf8(sc->menu.xftdraw, &sc->xftcolor[CWM_COLOR_MENU_FONT],
+	    sc->xftfont, 0, sc->xftfont->ascent + 1,
+	    (const FcChar8*)text, strlen(text));
+
+	free(text);
+}
