@@ -31,7 +31,7 @@
 
 #include "calmwm.h"
 
-#define MAXARGLEN 20
+static void	 log_msg(const char *, va_list);
 
 void
 u_spawn(char *argstr)
@@ -39,8 +39,7 @@ u_spawn(char *argstr)
 	switch (fork()) {
 	case 0:
 		u_exec(argstr);
-		err(1, "%s", argstr);
-		break;
+		exit(1);
 	case -1:
 		warn("fork");
 	default:
@@ -51,8 +50,10 @@ u_spawn(char *argstr)
 void
 u_exec(char *argstr)
 {
+#define MAXARGLEN 20
 	char	*args[MAXARGLEN], **ap = args;
 	char	**end = &args[MAXARGLEN - 1], *tmp;
+	char	*s = argstr;
 
 	while (ap < end && (*ap = strsep(&argstr, " \t")) != NULL) {
 		if (**ap == '\0')
@@ -75,8 +76,63 @@ u_exec(char *argstr)
 			}
 		}
 	}
-
 	*ap = NULL;
+
 	(void)setsid();
 	(void)execvp(args[0], args);
+	warn("%s", s);
+}
+
+char *
+u_argv(char * const *argv)
+{
+	size_t	 siz = 0;
+	int	 i;
+	char	*p;
+
+	if (argv == 0)
+		return(NULL);
+
+	for (i = 0; argv[i]; i++)
+		siz += strlen(argv[i]) + 1;
+	if (siz == 0)
+		return(NULL);
+
+	p = xmalloc(siz);
+	strlcpy(p, argv[0], siz);
+	for (i = 1; argv[i]; i++) {
+		strlcat(p, " ", siz);
+		strlcat(p, argv[i], siz);
+	}
+	return(p);
+}
+
+static void
+log_msg(const char *msg, va_list ap)
+{
+	char	*fmt;
+
+	if (asprintf(&fmt, "%s\n", msg) == -1) {
+		vfprintf(stderr, msg, ap);
+		fprintf(stderr, "\n");
+	} else {
+		vfprintf(stderr, fmt, ap);
+		free(fmt);
+	}
+	fflush(stderr);
+}
+
+void
+log_debug(int level, const char *func, const char *msg, ...)
+{
+	char	*fmt;
+	va_list	 ap;
+
+	if (Conf.debug < level)
+		return;
+
+	va_start(ap, msg);
+	xasprintf(&fmt, "debug%d: %s: %s", level, func, msg);
+	log_msg(fmt, ap);
+	va_end(ap);
 }
