@@ -77,11 +77,13 @@ xev_handle_maprequest(XEvent *ee)
 	XMapRequestEvent	*e = &ee->xmaprequest;
 	struct client_ctx	*cc = NULL, *old_cc;
 
-	if ((old_cc = client_current()))
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
+	if ((old_cc = client_current()) != NULL)
 		client_ptrsave(old_cc);
 
 	if ((cc = client_find(e->window)) == NULL)
-		cc = client_init(e->window, NULL);
+		cc = client_init(e->window, NULL, 0);
 
 	if ((cc != NULL) && (!(cc->flags & CLIENT_IGNORE)))
 		client_ptrwarp(cc);
@@ -93,6 +95,8 @@ xev_handle_unmapnotify(XEvent *ee)
 	XUnmapEvent		*e = &ee->xunmap;
 	struct client_ctx	*cc;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	if ((cc = client_find(e->window)) != NULL) {
 		if (e->send_event) {
 			client_set_wm_state(cc, WithdrawnState);
@@ -109,6 +113,8 @@ xev_handle_destroynotify(XEvent *ee)
 	XDestroyWindowEvent	*e = &ee->xdestroywindow;
 	struct client_ctx	*cc;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	if ((cc = client_find(e->window)) != NULL)
 		client_delete(cc);
 }
@@ -121,6 +127,8 @@ xev_handle_configurerequest(XEvent *ee)
 	struct screen_ctx	*sc;
 	XWindowChanges		 wc;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	if ((cc = client_find(e->window)) != NULL) {
 		sc = cc->sc;
 
@@ -174,6 +182,8 @@ xev_handle_propertynotify(XEvent *ee)
 	struct screen_ctx	*sc;
 	struct client_ctx	*cc;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	if ((cc = client_find(e->window)) != NULL) {
 		switch (e->atom) {
 		case XA_WM_NORMAL_HINTS:
@@ -209,67 +219,88 @@ xev_handle_enternotify(XEvent *ee)
 	XCrossingEvent		*e = &ee->xcrossing;
 	struct client_ctx	*cc;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	Last_Event_Time = e->time;
 
 	if ((cc = client_find(e->window)) != NULL)
 		client_setactive(cc);
 }
 
-/* We can split this into two event handlers. */
 static void
 xev_handle_buttonpress(XEvent *ee)
 {
 	XButtonEvent		*e = &ee->xbutton;
-	struct client_ctx	*cc, fakecc;
-	struct binding		*mb;
+	struct client_ctx	*cc;
+	struct screen_ctx	*sc;
+	struct bind_ctx		*mb;
+
+	LOG_DEBUG3("window: 0x%lx", e->window);
 
 	e->state &= ~IGNOREMODMASK;
 
-	TAILQ_FOREACH(mb, &Conf.mousebindingq, entry) {
+	TAILQ_FOREACH(mb, &Conf.mousebindq, entry) {
 		if (e->button == mb->press.button && e->state == mb->modmask)
 			break;
 	}
 
 	if (mb == NULL)
 		return;
-	if (mb->flags & CWM_WIN) {
+	mb->cargs->xev = CWM_XEV_BTN;
+	switch (mb->context) {
+	case CWM_CONTEXT_CC:
 		if (((cc = client_find(e->window)) == NULL) &&
 		    (cc = client_current()) == NULL)
 			return;
-	} else {
+		(*mb->callback)(cc, mb->cargs);
+		break;
+	case CWM_CONTEXT_SC:
 		if (e->window != e->root)
 			return;
-		cc = &fakecc;
-		cc->sc = screen_find(e->window);
+		if ((sc = screen_find(e->window)) == NULL)
+			return;
+		(*mb->callback)(sc, mb->cargs);
+		break;
+	case CWM_CONTEXT_NONE:
+		(*mb->callback)(NULL, mb->cargs);
+		break;
 	}
-
-	(*mb->callback)(cc, &mb->argument);
 }
 
 static void
 xev_handle_buttonrelease(XEvent *ee)
 {
-	struct client_ctx *cc;
+	XButtonEvent		*e = &ee->xbutton;
+	struct client_ctx	*cc;
 
-	if ((cc = client_current()))
-		group_toggle_membership_leave(cc);
+	LOG_DEBUG3("window: 0x%lx", ee->xbutton.window);
+
+	if ((cc = client_find(e->window)) != NULL) {
+		if (cc->flags & (CLIENT_ACTIVE | CLIENT_HIGHLIGHT)) {
+			cc->flags &= ~CLIENT_HIGHLIGHT;
+			client_draw_border(cc);
+		}
+	}
 }
 
 static void
 xev_handle_keypress(XEvent *ee)
 {
 	XKeyEvent		*e = &ee->xkey;
-	struct client_ctx	*cc = NULL, fakecc;
-	struct binding		*kb;
+	struct client_ctx	*cc;
+	struct screen_ctx	*sc;
+	struct bind_ctx		*kb;
 	KeySym			 keysym, skeysym;
 	unsigned int		 modshift;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	keysym = XkbKeycodeToKeysym(X_Dpy, e->keycode, 0, 0);
 	skeysym = XkbKeycodeToKeysym(X_Dpy, e->keycode, 0, 1);
 
 	e->state &= ~IGNOREMODMASK;
 
-	TAILQ_FOREACH(kb, &Conf.keybindingq, entry) {
+	TAILQ_FOREACH(kb, &Conf.keybindq, entry) {
 		if (keysym != kb->press.keysym && skeysym == kb->press.keysym)
 			modshift = ShiftMask;
 		else
@@ -278,22 +309,29 @@ xev_handle_keypress(XEvent *ee)
 		if ((kb->modmask | modshift) != e->state)
 			continue;
 
-		if (kb->press.keysym == (modshift == 0 ? keysym : skeysym))
+		if (kb->press.keysym == ((modshift == 0) ? keysym : skeysym))
 			break;
 	}
 
 	if (kb == NULL)
 		return;
-	if (kb->flags & CWM_WIN) {
+	kb->cargs->xev = CWM_XEV_KEY;
+	switch (kb->context) {
+	case CWM_CONTEXT_CC:
 		if (((cc = client_find(e->window)) == NULL) &&
 		    (cc = client_current()) == NULL)
 			return;
-	} else {
-		cc = &fakecc;
-		cc->sc = screen_find(e->window);
+		(*kb->callback)(cc, kb->cargs);
+		break;
+	case CWM_CONTEXT_SC:
+		if ((sc = screen_find(e->window)) == NULL)
+			return;
+		(*kb->callback)(sc, kb->cargs);
+		break;
+	case CWM_CONTEXT_NONE:
+		(*kb->callback)(NULL, kb->cargs);
+		break;
 	}
-
-	(*kb->callback)(cc, &kb->argument);
 }
 
 /*
@@ -304,15 +342,29 @@ xev_handle_keyrelease(XEvent *ee)
 {
 	XKeyEvent		*e = &ee->xkey;
 	struct screen_ctx	*sc;
+	struct client_ctx	*cc;
 	KeySym			 keysym;
 	unsigned int		 i;
 
-	sc = screen_find(e->root);
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
+	if ((sc = screen_find(e->root)) == NULL)
+		return;
 
 	keysym = XkbKeycodeToKeysym(X_Dpy, e->keycode, 0, 0);
 	for (i = 0; i < nitems(modkeys); i++) {
 		if (keysym == modkeys[i]) {
-			client_cycle_leave(sc);
+			if ((cc = client_current()) != NULL) {
+				if (sc->cycling) {
+					sc->cycling = 0;
+					client_mtf(cc);
+				}
+				if (cc->flags & CLIENT_HIGHLIGHT) {
+					cc->flags &= ~CLIENT_HIGHLIGHT;
+					client_draw_border(cc);
+				}
+			}
+			XUngrabKeyboard(X_Dpy, CurrentTime);
 			break;
 		}
 	}
@@ -325,42 +377,46 @@ xev_handle_clientmessage(XEvent *ee)
 	struct client_ctx	*cc, *old_cc;
 	struct screen_ctx       *sc;
 
-	sc = screen_find(e->window);
+	LOG_DEBUG3("window: 0x%lx", e->window);
 
-	if ((cc = client_find(e->window)) == NULL && e->window != sc->rootwin)
-		return;
-
-	if (e->message_type == cwmh[WM_CHANGE_STATE] && e->format == 32 &&
-	    e->data.l[0] == IconicState)
-		client_hide(cc);
-
-	if (e->message_type == ewmh[_NET_CLOSE_WINDOW])
-		client_send_delete(cc);
-
-	if (e->message_type == ewmh[_NET_ACTIVE_WINDOW] && e->format == 32) {
-		if ((old_cc = client_current()))
-			client_ptrsave(old_cc);
-		client_ptrwarp(cc);
+	if (e->message_type == cwmh[WM_CHANGE_STATE]) {
+		if ((cc = client_find(e->window)) != NULL) {
+	    		if (e->data.l[0] == IconicState)
+				client_hide(cc);
+		}
+	} else if (e->message_type == ewmh[_NET_CLOSE_WINDOW]) {
+		if ((cc = client_find(e->window)) != NULL) {
+			client_send_delete(cc);
+		}
+	} else if (e->message_type == ewmh[_NET_ACTIVE_WINDOW]) {
+		if ((cc = client_find(e->window)) != NULL) {
+			if ((old_cc = client_current()) != NULL)
+				client_ptrsave(old_cc);
+			client_show(cc);
+			client_ptrwarp(cc);
+		}
+	} else if (e->message_type == ewmh[_NET_WM_DESKTOP]) {
+		if ((cc = client_find(e->window)) != NULL) {
+			/*
+			 * The EWMH spec states that if the cardinal returned
+			 * is 0xFFFFFFFF (-1) then the window should appear
+			 * on all desktops, in our case, group 0.
+			 */
+			if (e->data.l[0] == (unsigned long)-1)
+				group_movetogroup(cc, 0);
+			else
+				group_movetogroup(cc, e->data.l[0]);
+		}
+	} else if (e->message_type == ewmh[_NET_WM_STATE]) {
+		if ((cc = client_find(e->window)) != NULL) {
+			xu_ewmh_handle_net_wm_state_msg(cc,
+			    e->data.l[0], e->data.l[1], e->data.l[2]);
+		}
+	} else if (e->message_type == ewmh[_NET_CURRENT_DESKTOP]) {
+		if ((sc = screen_find(e->window)) != NULL) {
+			group_only(sc, e->data.l[0]);
+		}
 	}
-
-	if (e->message_type == ewmh[_NET_WM_DESKTOP] && e->format == 32) {
-		/*
-		 * The EWMH spec states that if the cardinal returned is
-		 * 0xFFFFFFFF (-1) then the window should appear on all
-		 * desktops, which in our case is assigned to group 0.
-		 */
-		if (e->data.l[0] == (unsigned long)-1)
-			group_movetogroup(cc, 0);
-		else
-			group_movetogroup(cc, e->data.l[0]);
-	}
-
-	if (e->message_type == ewmh[_NET_WM_STATE] && e->format == 32)
-		xu_ewmh_handle_net_wm_state_msg(cc,
-		    e->data.l[0], e->data.l[1], e->data.l[2]);
-
-	if (e->message_type == ewmh[_NET_CURRENT_DESKTOP] && e->format == 32)
-		group_only(sc, e->data.l[0]);
 }
 
 static void
@@ -370,11 +426,14 @@ xev_handle_randr(XEvent *ee)
 	struct screen_ctx		*sc;
 	int				 i;
 
+	LOG_DEBUG3("new size: %d/%d", rev->width, rev->height);
+
 	i = XRRRootToScreen(X_Dpy, rev->root);
 	TAILQ_FOREACH(sc, &Screenq, entry) {
 		if (sc->which == i) {
 			XRRUpdateConfiguration(ee);
 			screen_update_geometry(sc);
+			screen_assert_clients_within(sc);
 		}
 	}
 }
@@ -389,6 +448,8 @@ xev_handle_mappingnotify(XEvent *ee)
 	XMappingEvent		*e = &ee->xmapping;
 	struct screen_ctx	*sc;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	XRefreshKeyboardMapping(e);
 	if (e->request == MappingKeyboard) {
 		TAILQ_FOREACH(sc, &Screenq, entry)
@@ -402,6 +463,8 @@ xev_handle_expose(XEvent *ee)
 	XExposeEvent		*e = &ee->xexpose;
 	struct client_ctx	*cc;
 
+	LOG_DEBUG3("window: 0x%lx", e->window);
+
 	if ((cc = client_find(e->window)) != NULL && e->count == 0)
 		client_draw_border(cc);
 }
@@ -411,9 +474,11 @@ xev_process(void)
 {
 	XEvent		 e;
 
-	XNextEvent(X_Dpy, &e);
-	if (e.type - Randr_ev == RRScreenChangeNotify)
-		xev_handle_randr(&e);
-	else if (e.type < LASTEvent && xev_handlers[e.type] != NULL)
-		(*xev_handlers[e.type])(&e);
+	while (XPending(X_Dpy)) {
+		XNextEvent(X_Dpy, &e);
+		if (e.type - Conf.xrandr_event_base == RRScreenChangeNotify)
+			xev_handle_randr(&e);
+		else if (e.type < LASTEvent && xev_handlers[e.type] != NULL)
+			(*xev_handlers[e.type])(&e);
+	}
 }
